import os
import shutil
import uuid
import json as jn
from loguru import logger
import re
from equator_qa.prompts import warning_prompt, scoring_criteria
from equator_qa.llmscalls import (
    call_ollama_evaluator_api,
    call_groq_evaluator_api,
    call_openrouter_student_api,
    call_ollama_student_api,
    call_groq_student_api,
)

import sqlite3

def begin_benchmark(
    execution_steps,
    student_ollama_models,
    student_groq_models,
    student_openrouter_models,
    OLLAMA_EVALUATOR_MODEL,
    GROQ_EVALUATOR_MODEL,
    vectordb_instance,
    benchmark_name,
    date_now,
    answer_rounds,
):
    print("line 30, UTILS, Starting benchmark execution steps =", execution_steps)

    if "ollama_to_groq_evaluate" in execution_steps:
        print("ollama_to_groq_evaluate == execution_steps", execution_steps)
        for model in student_groq_models:
            student_model = model
            evaluator_model = OLLAMA_EVALUATOR_MODEL
            client = EQUATOR_Client(
                execution_steps,
                student_model,
                evaluator_model,
                vectordb_instance,
            )
            lab = "equator"  # TODO: think about a more generic way of doing this
            if student_model:
                print(f"Extracted Lab name: {lab}")
                print(f"Extracted model name: {student_model}")
            else:
                print("Model name not found.")

            student_models = [student_model]
            print("1. GETTING EQUATOR Evaluator ANSWERS -Local Student")

            model_path = ""
            folder_name = f"{date_now}-{benchmark_name}"
            # answers_save_path = f"./{folder_name}/llm_outputs"
            auto_eval_save_path = f"./{folder_name}/auto_eval_outputs"
            # stats_save_path = f"./{folder_name}/tables_and_charts"
            for n in range(answer_rounds):
                print(f"\n----- Round: {n+1} of {answer_rounds} -----")
                answer_save_path_round = f"{auto_eval_save_path}"
                client.EQUATOR_Controller(
                    model_path,
                    lab,
                    student_models,
                    answer_save_path_round=answer_save_path_round,
                    count=n,
                    prefix_replace="auto_eval-",
                )

    elif "groq_to_ollama_evaluate" in execution_steps:
        print("groq_to_ollama_evaluate == execution_steps", execution_steps)

        for model in student_ollama_models:
            evaluator_model = GROQ_EVALUATOR_MODEL
            student_model = model
            client = EQUATOR_Client(
                execution_steps,
                student_model,
                evaluator_model,
                vectordb_instance,
            )
            lab = "equator"
            if student_model:
                print(f"Extracted Lab name: {lab}")
                print(f"Extracted model name: {student_model}")
            else:
                print("Model name not found.")
            student_models = [student_model]
            print("1. GETTING EQUATOR Evaluator ANSWERS -Local Student")
            model_path = ""
            folder_name = f"{date_now}-{benchmark_name}"
            # answers_save_path = f"./{folder_name}/llm_outputs"
            auto_eval_save_path = f"./{folder_name}/auto_eval_outputs"
            # stats_save_path = f"./{folder_name}/tables_and_charts"
            for n in range(answer_rounds):
                print(f"\n----- Round: {n+1} of {answer_rounds} -----")
                answer_save_path_round = f"{auto_eval_save_path}"
                client.EQUATOR_Controller(
                    model_path,
                    lab,
                    student_models,
                    answer_save_path_round=answer_save_path_round,
                    count=n,
                    prefix_replace="auto_eval-",
                )
    if "ollama_to_openrouter_evaluate" in execution_steps:
        print("ollama_to_openrouter_evaluate == execution_steps", execution_steps)

        for model in student_openrouter_models:
            model_path = model
            evaluator_model = OLLAMA_EVALUATOR_MODEL
            lab, student_model = extract_model_parts(model)
            lab = "equator"
            if student_model:
                print(f"Extracted Lab name: {lab}")
                print(f"Extracted model name: {student_model}")
            else:
                print("Model name not found.")
            evaluator_model = OLLAMA_EVALUATOR_MODEL
            student_models = [student_model]
            evaluator_model = model_path
            client = EQUATOR_Client(
                execution_steps,
                student_model,
                evaluator_model,
                vectordb_instance,
            )
            folder_name = f"{date_now}-{benchmark_name}"
            # answers_save_path = f"./{folder_name}/llm_outputs"
            auto_eval_save_path = f"./{folder_name}/auto_eval_outputs"
            # stats_save_path = f"./{folder_name}/tables_and_charts"
            print("1. GETTING BERNARD LLM Evaluator ANSWERS")
            for n in range(answer_rounds):
                print(f"\n----- Round: {n+1} of {answer_rounds} -----")
                answer_save_path_round = f"{auto_eval_save_path}"
                client.EQUATOR_Controller(
                    model_path,
                    lab,
                    student_models,
                    answer_save_path_round=answer_save_path_round,
                    count=n,
                    prefix_replace="auto_eval-",
                )
    if "groq_to_openrouter_evaluate" in execution_steps:
        print("groq_to_openrouter_evaluate == execution_steps", execution_steps)
        for model in student_openrouter_models:
            model_path = model
            student_model = model
            lab = "equator"
            if student_model:
                print(f"Extracted Lab name: {lab}")
                print(f"Extracted model name: {student_model}")
            else:
                print("Model name not found.")
            student_model = student_model.replace('/', "-")

            student_models = [student_model]
            
            evaluator_model = GROQ_EVALUATOR_MODEL
            client = EQUATOR_Client(
                execution_steps,
                student_model,
                evaluator_model,
                vectordb_instance,
            )
            folder_name = f"{date_now}-{benchmark_name}"
            # answers_save_path = f"./{folder_name}/llm_outputs"
            auto_eval_save_path = f"./{folder_name}/auto_eval_outputs"
            # stats_save_path = f"./{folder_name}/tables_and_charts"
            print("1. GETTING BERNARD LLM Evaluator ANSWERS")
            for n in range(answer_rounds):
                print(f"\n----- Round: {n+1} of {answer_rounds} -----")
                answer_save_path_round = f"{auto_eval_save_path}"
                client.EQUATOR_Controller(
                    model_path,
                    lab,
                    student_models,
                    answer_save_path_round=answer_save_path_round,
                    count=n,
                    prefix_replace="auto_eval-",
                )

class EQUATOR_Client(object):
    def __init__(
        self,
        execution_steps,
        student_model,
        evaluator_model,
        vectordb_instance,
    ):
        self.student_model = student_model
        self.evaluator_model = evaluator_model
        self.execution_steps = execution_steps
        self.vectordb2 = vectordb_instance

    def EQUATOR_Controller(
        self,
        model_path,
        lab,
        student_models,
        answer_save_path_round,
        count,
        prefix_replace,
    ):
        print("prefix ==", prefix_replace)

        ## Go through the vector db chroma.sqlite3 for all the question and id
        # Path to your chroma.sqlite3 file
        db_path = "chroma.sqlite3"
        # Connect to the SQLite database
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        batch_size = 100  # Define your batch size
        offset = 0  # Start offset
        stop_processing = False  # Flag to stop the outer loop
        while True:
            # Fetch a batch of results
            query = f"""
            SELECT 
                json_extract(string_value, '$.id'), 
                json_extract(string_value, '$.category'), 
                json_extract(string_value, '$.question'), 
                json_extract(string_value, '$.response') 
            FROM embedding_fulltext_search
            LIMIT {batch_size} OFFSET {offset}
            """
            print(f"Executing query with OFFSET={offset}, LIMIT={batch_size}")
            cursor.execute(query)
            results = cursor.fetchall()

            # Break the loop if no more records are fetched
            if not results:
                print("No more records found. Exiting.")
                break

            for row in results:
                question_id, category, question_text, response = row
                print(
                    f"Processing ID: {question_id}, Category: {category}, Question: {question_text}, Answer: {response}"
                )

                # Simulate stripping and processing text
                question = question_text.strip() if question_text else ""
                human_answer = response.strip() if response else ""

                for student_model in student_models:
                    output_path = f"{answer_save_path_round}/round_{count + 1}/{prefix_replace}{lab+"-"}{student_model}.json"
                    print("line 105 Model Path  = ", model_path)

                    # Call your evaluator function
                    student_answer, evaluator_response = self.call_evaluator(
                        model_path=model_path,
                        prompt=question,
                    )

                    print("line 256, unpacking student answer = ", student_answer)
                    print("line 257, unpacking evaluator  = ", evaluator_response)

                    score = extract_score_from_string(evaluator_response)

                    create_template_json(
                        student_model,
                        output_path,
                        question_id,
                        category,
                        human_answer,
                        question,
                        student_answer,
                        evaluator_response,
                        score,
                    )
                # Stop processing if a condition is met
                if question_id == "1":  # Replace "1" with the desired stop condition
                    print("Stop condition met. Exiting.")
                    stop_processing = True  # Set the flag to stop outer loop
                    break
            # Increment offset to fetch the next batch
            if stop_processing:
                print("Breaking the outer loop.")
                break
            offset += batch_size  # Move to the next batch

        # Close the database connection after processing
        conn.close()
        print("Database connection closed.")

    # Generate student answer
    def student(self, model_path, full_prompt_student):
        model_path = str(model_path)
        # print("line 442 - Model Path = ", model_path)
        print("line 260 ,UTILS, in Execution steps =", self.execution_steps)
        if "ollama_to_openrouter_evaluate" in self.execution_steps:
            print("line 293, UTILS , in Execution steps =", self.execution_steps)

            response = call_openrouter_student_api(
                full_prompt_student, warning_prompt, model_path
            )
            return response

        elif "groq_to_ollama_evaluate" in self.execution_steps:
            print("line 301, UTILS , in Execution steps =", self.execution_steps)

            response = call_ollama_student_api(
                full_prompt_student, warning_prompt, self.student_model
            )
            return response
        elif "ollama_to_groq_evaluate" in self.execution_steps:
            print("line 308, UTILS , in Execution steps =", self.execution_steps)

            response = call_groq_student_api(
                full_prompt_student, warning_prompt, self.student_model
            )
            print("line 180, def student() = response ", response)
            return response

        elif "groq_to_openrouter_evaluate" in self.execution_steps:
            print("line 317, UTILS , in Execution steps =", self.execution_steps)

            response = call_openrouter_student_api(
                full_prompt_student, warning_prompt, model_path
            )
            print("line 288, UTILS,  = response ", response)
            return response

    def call_evaluator(self, model_path, prompt):
        results = self.vectordb2.retrieve_embedding(prompt)
        if results is None:
            logger.error("Failed to retrieve similar documents.")
            return None
        context = ""
        if "documents" in results and results["documents"]:
            metadatas = results.get("metadatas", [])[0]
            for metadata in metadatas:
                context += f"Question: {metadata.get('question', '')}\n"
                context += f"Answer: {metadata.get('response', '')}\n\n"
            logger.info(context)
        else:
            logger.warning("No similar documents found.")

        student_answer = self.student(model_path, prompt)
        print("line 341, UTILS, student_answer =", student_answer)
        if not student_answer:
            logger.error("Failed to get Student Answer.")
            return None

        logger.info(f"Student Answer: {student_answer}")

        evaluator_system_prompt = [
            {
                "role": "system",
                "content": (
                    "You are a strict high school teacher. "
                    f"Provide the score in the form of a JSON object: {{'score': <0 or 100>}}"
                    "Your role is to grade strictly based on the provided answer key. "
                    "Only examine the final answer given by the student in a JSON object: {'student_answer': <My final Answer>}. "
                    "If the student's answer matches the answer key, assign a score of 100; otherwise, assign a score of 0. "
                    "No partial credit allowed! Do not consider logical arguments or reasoning; they are irrelevant in this case. "
                    f"Provide the score in the form of a JSON object: {{'score': <0 or 100>}}."
                ),
            },
            {
                "role": "user",
                "content": (
                    f"Scoring Criteria: {sanitize_string(scoring_criteria)}\n"
                    f"Answer key: {sanitize_string(context)}\n"
                    f"Student answer: {sanitize_string(student_answer)}"
                ),
            },
        ]

        if "ollama_to_groq_evaluate" in self.execution_steps:

            student_answer, eval_response = call_ollama_evaluator_api(
                self.evaluator_model, student_answer, evaluator_system_prompt
            )
            return student_answer, eval_response

        elif "groq_to_ollama_evaluate" or "groq_to_openrouter_evaluate" in self.execution_steps:
            student_answer, eval_response = call_groq_evaluator_api(
                self.evaluator_model, student_answer, evaluator_system_prompt
            )
            return student_answer, eval_response

        elif "ollama_to_openrouter_evaluate" in self.execution_steps:
            student_answer, eval_response = call_ollama_evaluator_api(
                self.evaluator_model, student_answer, evaluator_system_prompt
            )
            return student_answer, eval_response
        elif "groq_to_openrouter_evaluate" in self.execution_steps:
            student_answer, eval_response = call_groq_evaluator_api(
                self.evaluator_model, student_answer, evaluator_system_prompt
            )
            return student_answer, eval_response



def extract_model_parts(model_string):
    # Define the regex pattern to extract both parts
    pattern = r"^([^/]+)/([^/:]+)"
    # Use re.match to find the model parts
    match = re.match(pattern, model_string)
    if match:
        return match.group(1), match.group(2)
    return None, None


def sanitize_string(value):
    """
    Escapes curly braces in strings to prevent issues with format specifiers in logging.
    """
    if isinstance(value, str):
        return value.replace("{", "{{").replace("}", "}}")
    return value


def create_template_json(
    student_model,
    output_path,
    question_id,
    category,
    human_answer,
    question,
    student_answer,
    evaluator_response,
    score,
):
    # Ensure the directory for the output path exists
    output_dir = os.path.dirname(output_path)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    logger.info(student_model)
    # Load existing data if the file exists
    if os.path.exists(output_path):
        try:
            with open(output_path, "r", encoding="utf-8") as infile:
                template_data = jn.load(infile)
        except (jn.JSONDecodeError, FileNotFoundError):
            template_data = {}  # Start fresh if file is empty or corrupted
    else:
        template_data = {}

    # Define or update the structure of the template JSON
    template_data[question_id] = {
        "category": category,
        "question": question,
        "human_answer": human_answer,
        "model_answer": student_answer,
        "eval_response": evaluator_response,
        "score": score,
    }

    # Write the updated data back to the file
    with open(output_path, "w", encoding="utf-8") as json_file:
        jn.dump(template_data, json_file, indent=4, ensure_ascii=False)

    print(f"Template JSON created/updated: {output_path}")


def extract_score_from_string(response_string):
    # Regular expressions to match different patterns that indicate a score
    patterns = [
        r"\"score\"\s*:\s*(\d+)",  # JSON-like: "score": 0 or "score":0
        r"'score':\s*(\d+)",  # Python dict-like: {'score': 0}
        r"'grade':\s*(\d+)",  # Python dict-like: {'grade': 0}
        r"Grade:\s*(\d+)",  # Grade without ratio, e.g., Grade: 0
        r"Grade:\s*{'score':\s*(\d+)}",  # Grade followed by Python dict, e.g., Grade: {'score': 0}
        r"Score:\s*{'score':\s*(\d+)}",  # Score followed by Python dict, e.g., Score: {'score': 0}
        r"\*\*Score:\*\*\s*{'score':\s*(\d+)}",  # Markdown Score followed by Python dict, e.g., **Score:** {'score': 20}
        r"\*\*Grade:\*\*\s*{'score':\s*(\d+)}",  # Markdown Grade followed by Python dict, e.g., **Grade:** {'score': 0}
        r"score\s*is\s*(\d+)%",  # Plain text: score is 0%
        r"score\s*of\s*\*\*(\d+)%\*\*",  # Markdown: score of **0%**
        r"the\s*score\s*assigned\s*is\s*(\d+)%",  # Assigned score: the score assigned is 0%
        r"Grade:\s*A\s*\(\s*(\d+)%\)",  # Grade with percentage, e.g., Grade: A (100%)
        r"Grade:\s*[F]\s*\(\s*(\d+)/\d+\)",  # Grade F with ratio, e.g., Grade: F (0/10)
        r"Grade:\s*(\d+)/\d+",  # Ratio format, e.g., Grade: 0/10
        r"\*\*Grade:\*\*\s*(\d+)/\d+",  # Markdown style: **Grade:** 0/10
        r"\*\*Grade:\*\*\s*F\s*\(\s*(\d+)/\d+\)",  # Markdown style with grade F: **Grade:** F (0/100)
        r"Grade:\s*\*\*(\d+)/\d+\*\*",  # Markdown format, e.g., **Grade:** 0/10
        r"Grade:\s*F\s*\(\s*(\d+)\s*out\s*of\s*\d+\)",  # Grade F with "out of", e.g., Grade: F (0 out of 10)
        r"You\s*received\s*a\s*score\s*of\s*(\d+)\s*out\s*of\s*\d+",  # Plain text: You received a score of 0 out of 10
        r"\*\*(\d+)/100\s*score\*\*",  # Markdown style, e.g., **100/100 score**
        r"would\s*earn\s*a\s*score\s*of\s*(\d+)",  # Plain text: would earn a score of 100
        r"return\s*a\s*score\s*of\s*(\d+)",  # Plain text: return a score of 0
    ]

    # Iterate over each pattern to find a match
    for pattern in patterns:
        match = re.search(pattern, response_string, re.IGNORECASE)
        if match:
            return int(match.group(1))

    # If no matching score pattern is found, return None
    return None


# def unload_model(model):
#     keep_alive = 0
#     url = "http://localhost:11434/api/generate"
#     payload = json.dumps({
#         "model": model,
#         "keep_alive": keep_alive
#     })
#     headers = {
#         "Content-Type": "application/json"
#     }

#     try:
#         response = requests.post(url, data=payload, headers=headers)
#         response.raise_for_status()
#         return response.json()
#     except requests.exceptions.RequestException as e:
#         return {"error": str(e)}


def is_valid_uuid4(name: str) -> bool:
    """Return True if 'name' is a valid UUID version 4 string."""
    try:
        val = uuid.UUID(name, version=4)
    except ValueError:
        return False
    return str(val) == name


def cleanup_chromadb(db_filename="chroma.sqlite3"):
    # 1. Remove chromadb.sqlite3 file if it exists
    os.remove(db_filename)

    # 3. Remove directories in the current directory that look like UUIDv4
    for item in os.listdir("."):
        if os.path.isdir(item) and is_valid_uuid4(item):
            print(f"Removing UUID-like directory: {item}")
            shutil.rmtree(item)



import os 
import pandas as pd
import numpy as np


def load_all_llm_answers_from_json(
    answers_save_path: str,
    prefix_replace="auto-eval",
    sub_folders=[""],
) -> list[dict[pd.DataFrame]]:
    # reload all the scored answers from json files
    all_llm_answers = {}
    for sub_folder in sub_folders:
        answers_save_path_sub = f"{answers_save_path}{sub_folder}"
        if not os.path.exists(answers_save_path_sub):
            continue
        for output_file in os.listdir(f"{answers_save_path_sub}/"):
            if output_file.endswith(".json"):
                outputs_df = pd.read_json(
                    f"{answers_save_path_sub}/{output_file}", orient="index"
                )
                model = output_file.replace(prefix_replace, "").replace(".json", "")
                all_llm_answers.setdefault(model, pd.DataFrame())
                all_llm_answers[model] = pd.concat([all_llm_answers[model], outputs_df])
    print("test statistics, all answers", all_llm_answers)
    return all_llm_answers


def get_llm_stats(
    all_llm_answers: dict, stats_save_path: str, file_suffix="", bootstrap_n=10000
) -> pd.DataFrame:
    all_llm_stats = calculate_llm_stats(all_llm_answers, bootstrap_n)
    stats_df = (
        pd.DataFrame(all_llm_stats)
        .transpose()
        .sort_values("mean_score", ascending=False)
    )
    stats_df.index.name = "model"
    os.makedirs(stats_save_path, exist_ok=True)
    stats_df.to_csv(f"./{stats_save_path}/final_stats{file_suffix}.csv")
    return stats_df


def calculate_llm_stats(all_llm_answers: dict, bootstrap_n=10000) -> dict:
    all_llm_stats = {}
    for model, outputs in all_llm_answers.items():
        print(f"Calculating stats for {model}")
        mean_score = outputs["score"].mean()
        std_dev_score = outputs["score"].std()
        # do a n(10,000) bootstrap to get the 95% CI
        bootstrap_scores = []
        for _ in range(bootstrap_n):
            bootstrap_scores.append(
                outputs["score"].sample(frac=1, replace=True).mean()
            )
        ci_lower = np.percentile(bootstrap_scores, 2.5)
        ci_upper = np.percentile(bootstrap_scores, 97.5)
        # caculate z-interval 95%
        z = 1.96
        z_interval_error = z * (std_dev_score / np.sqrt(len(outputs)))
        all_llm_stats[model] = {
            "mean_score": mean_score,
            "std_dev_score": std_dev_score,
            "z_interval_error": z_interval_error,
            "ci_lower": ci_lower,
            "ci_upper": ci_upper,
            "output_count": len(outputs),
        }
    return all_llm_stats




scoring_criteria = r"""SCORING CRITERIA
100%: The response must match the answer key given, even if the student used bad reasoning, logic to arrive at the final answer. 
0%: The does NOT match the answer key given. No partial credit allowed!
You must only return a JSON object with the score {'score': <0 or 100>}

TASK

Evaluate whether the STUDENT Answer matches the answer key given. If it does, assign a score of 100% Otherwise you must assign a score of 0%.  Provide a very short explanation on why.  
Just focus on the student's final answer!  Give full credit to the student if final answer matches answer key. Don't over think this.  Also do not evaluate based on the quality, logical reasoning , even if it is very persuasive!
Only consider the answer key as the source of true.  Your job is at risk of you do not follow our instructions.  If it the Answer Key match the students answer you must assign a score of 0%, no partial credit allowed.    
Return a JSON object with the explanation of why the student got his\her score == {"score": <0 or 100>}.  Keep it do less than two sentences {"evaluation": "<explanation>"}
You must only return a JSON object with the score {'score': <0 or 100>}
No partial credit allowed!"""

warning_prompt = r"""You think very carefully about the question asked. You make zero assumptions about classic problems. You are not to be tricked! 
Warning: THIS IS NOT THE CLASSIC PROBLEM! \n"""




from loguru import logger
import json
import chromadb
import requests


class VectorDB_Controller(object):
    def __init__(self, keepVectorDB):
        self.base_url = "http://localhost:11434"  # in container
        self.chroma_client = chromadb.PersistentClient(path=".")
        self.keepVectorDB = keepVectorDB
        if not keepVectorDB:
            # open file for vector score
            # Open and load the JSON file
            with open("data/linguistic_benchmark.json", "r", encoding="utf-8") as f:
                data = json.load(f)

            # Initialize a list to store the parsed conversations
            conversations = []

            # Iterate over the list of dictionaries
            for entry in data:
                parsed_entry = {
                    "id": entry.get(
                        "index", ""
                    ),  # Use "index" as the unique identifier
                    "category": entry.get("category", ""),  # Extract category
                    "question": entry.get("question", ""),  # Extract the question
                    "response": entry.get(
                        "human_answer", ""
                    ),  # Extract the human answer
                }
                conversations.append(parsed_entry)
            logger.info(conversations)
            self.create_vector_db(conversations)

    def generate_embeddings(self, model, input_text, truncate=True):
        url = f"{self.base_url}/api/embed"
        payload = {"model": model, "input": input_text, "truncate": truncate}
        try:
            response = requests.post(url, json=payload)
            response.raise_for_status()  # Raise an error for bad responses
            if response.headers.get("Content-Type").startswith("application/json"):
                response_json = response.json()
                if "embeddings" in response_json:
                    return response_json
                else:
                    logger.error(f"No embeddings found in response: {response_json}")
                    return None
            else:
                logger.error(
                    f"Unexpected response content type: {response.headers.get('Content-Type')}"
                )
                logger.error(f"Response content: {response.text}")
                return None
        except requests.RequestException as e:
            logger.error(f"Failed to generate embeddings: {e}")
            return None

    def retrieve_embedding(self, prompt, n_results=1):
        # print("line 411 retrieve embedding prompt  ==", prompt)
        response = self.generate_embeddings(model="all-minilm", input_text=prompt)
        # print("line 413 Generate Embeddings ==", response)
        if not response or "embeddings" not in response:
            logger.error("Failed to retrieve embeddings from the model.")
            return None
        prompt_embedding = self.flatten_embedding(response["embeddings"])
        vector_db = self.chroma_client.get_collection(name="conversations")
        try:
            results = vector_db.query(
                query_embeddings=[prompt_embedding], n_results=n_results
            )
            return results
        except Exception as e:
            logger.error(f"Error querying vector DB: {e}")
            return None

    def add_to_vector_db(self, vector_db, entry_id, serialized_conversations, metadata):
        response = self.generate_embeddings(
            model="all-minilm", input_text=serialized_conversations
        )

        if not response or "embeddings" not in response:
            logger.error(
                f"Failed to retrieve embeddings for entry {entry_id}. Response: {response}"
            )
            return
        # Flatten the embedding if it is nested
        embedding = self.flatten_embedding(response["embeddings"])

        converted_metadata = {
            k: self.convert_metadata_value(v) for k, v in metadata.items()
        }

        try:
            vector_db.add(
                ids=[str(entry_id)],
                embeddings=[embedding],
                documents=[serialized_conversations],
                metadatas=[converted_metadata],
            )
        except Exception as e:
            logger.error(f"Error adding entry {entry_id} to the vector DB: {e}")

    def create_vector_db(self, conversations):
        vector_db_name = "conversations"
        try:
            self.chroma_client.delete_collection(name=vector_db_name)
        except ValueError:
            pass  # Handle collection not existing
        vector_db = self.chroma_client.create_collection(name=vector_db_name)
        for c in conversations:
            serialized_conversations = json.dumps(c)
            self.add_to_vector_db(vector_db, c["id"], serialized_conversations, c)

    def convert_metadata_value(self, value):
        if value is None:
            return ""
        if isinstance(value, (list, dict)):
            return json.dumps(value)
        return value

    def flatten_embedding(self, embedding):
        # Flatten nested embeddings if necessary
        if isinstance(embedding[0], list):
            return [item for sublist in embedding for item in sublist]
        return embedding

    def add_to_vector_db(self, vector_db, entry_id, serialized_conversations, metadata):
        response = self.generate_embeddings(
            model="all-minilm", input_text=serialized_conversations
        )
        if not response or "embeddings" not in response:
            logger.error(
                f"Failed to retrieve embeddings for entry {entry_id}. Response: {response}"
            )
            return
        # Flatten the embedding if it is nested
        embedding = self.flatten_embedding(response["embeddings"])

        converted_metadata = {
            k: self.convert_metadata_value(v) for k, v in metadata.items()
        }

        try:
            vector_db.add(
                ids=[str(entry_id)],
                embeddings=[embedding],
                documents=[serialized_conversations],
                metadatas=[converted_metadata],
            )
        except Exception as e:
            logger.error(f"Error adding entry {entry_id} to the vector DB: {e}")



import os               
import json
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import numpy as np
from IPython.display import display
from equator.stats import load_all_llm_answers_from_json, get_llm_stats

mapper = {
    'gpt-4-turbo-preview': 'GPT-4 Turbo',
    'gpt-4o': 'GPT-4o',
    'gpt-4o-mini-2024-07-18': 'GPT-4o Mini',
    'claude-3-opus-20240229': 'Claude 3 Opus',
    'claude-3-5-sonnet-20240620': 'Claude 3.5 Sonnet',
    'gemini-1_5-pro': 'Gemini 1.5 Pro',
    'gemini-1_0-pro': 'Gemini 1.0 Pro',
    'gemini-1_5-pro-exp-0801': 'Gemini 1.5 Pro Ex',
    'mistral-large-latest': 'Mistral Large 2',
    'open-mixtral-8x22b': 'Mistral 8x22B',
    'meta_llama3-70b-instruct-v1_0': 'Llama 3 70B',
    'meta_llama3-1-70b-instruct-v1_0': 'Llama 3.1 70B',
    'command-r': 'Command R',
    'command-r-plus': 'Command R Pro',
    'Meta-Llama-3-1-405B-Instruct-jjo_eastus_models_ai_azure_com': 'Llama 3.1 405B',
    'Meta-Llama-3-1-70B-Instruct-ostu_eastus_models_ai_azure_com': 'Llama 3.1 70B',
}

def define_data(final_stats: pd.DataFrame):
    ## Define the data
    # models = ["Human level*", "GPT-4 Turbo", "Claude 3 Opus", "Mistral Large", "Gemini Pro 1.5",
    #           "Gemini Pro 1.0", "Llama 3 70B", "Mistral 8x22B"]
    # mean_scores = [80, 38, 33, 30, 29, 27, 21, 16]
    # lower_bounds = [10, 16, 15, 15, 15, 15, 13, 11]
    # upper_bounds = [10, 16, 15, 15, 15, 15, 13, 11]

    final_stats['model'] = final_stats['model'].map(mapper).fillna(final_stats['model'])
    final_stats.loc[-1] = {
        'model': 'Human level*',
        'mean_score': 86,
        'std_dev_score': 0,
        'z_interval_error': 0,
        'ci_lower': 93,
        'ci_upper': 78,
    }
    final_stats = final_stats.sort_values(by='mean_score', ascending=False)

    models = final_stats['model'].to_list()
    mean_scores = final_stats['mean_score'].to_list()
    lower_bounds = final_stats['ci_lower'].to_list()
    upper_bounds = final_stats['ci_upper'].to_list()

    data = {
        "Model": models,
        "Average": mean_scores,
        "Confidence Interval Low": lower_bounds,
        "Confidence Interval High": upper_bounds,
    }
    return pd.DataFrame(data)


def create_performance_chart(final_stats: pd.DataFrame, title="LLM Linguistic Benchmark Performance",
                             highlight_models=None):
    if highlight_models is None:
        highlight_models = []

    df = define_data(final_stats)
    # Create a basic barplot
    sns.set_theme(style="whitegrid")
    plt.figure(figsize=(10, 6))

    # Different colors for different models
    colors = ['skyblue' if model not in highlight_models else 'orange' for model in df["Model"]]
    barplot = sns.barplot(data=df, x="Model", y="Average", palette=colors, errorbar=None)

    # Shade the first bar with black cross lines
    for i, bar in enumerate(barplot.patches):  # Loop through the bars
        if df["Model"][i] == "Human level*":  # Check if it is the Human level bar
            bar.set_hatch('///')  # Apply hatching

    # Add confidence intervals as vertical lines with caps
    capwidth = 0.1  # Width of the cap lines
    for i, model in enumerate(df["Model"]):
        plt.plot([i, i], [df["Confidence Interval Low"][i], df["Confidence Interval High"][i]],
                 color='grey', lw=1)
        # Add horizontal caps
        plt.plot([i - capwidth / 2, i + capwidth / 2],
                 [df["Confidence Interval Low"][i], df["Confidence Interval Low"][i]],
                 color='grey', lw=1)
        plt.plot([i - capwidth / 2, i + capwidth / 2],
                 [df["Confidence Interval High"][i], df["Confidence Interval High"][i]],
                 color='grey', lw=1)

    plt.title(title, fontsize=18)
    plt.xlabel("", fontsize=14)
    plt.ylabel("Average Score (%)", fontsize=14)
    plt.xticks(rotation=60, fontsize=14)
    plt.tight_layout()

    return barplot, plt

def create_basic_charts(execution_steps,answer_rounds,benchmark_name,date_now):
     if "generate_statistics" in execution_steps:
        folder_name = f"{date_now}-{benchmark_name}"
        auto_eval_save_path = f"./{folder_name}/auto_eval_outputs"
        stats_save_path = f"./{folder_name}/tables_and_charts"
        sub_eval_folders = [f"/round_{r+1}" for r in range(answer_rounds)]
        print("2. GENERATING STATISTICS")
        all_stats_dfs = {}
        save_info = [
            {
                "path": auto_eval_save_path,
                "chart_title": "LLM Linguistic Benchmark Performance",
                "type": "",
            }
        ]
        for info in save_info:
            save_path = info["path"]
            chart_title = info["chart_title"]
            info_type = info["type"]
            print("line 120 CHARTING, Eval for path:", save_path)
            all_llm_evals = load_all_llm_answers_from_json(
                save_path,
                prefix_replace="auto_eval-",
                sub_folders=sub_eval_folders,
            )
            stats_df = get_llm_stats(
                all_llm_evals, stats_save_path, file_suffix=info_type, bootstrap_n=10000
            )
            display(stats_df)
            barplot, plt = create_performance_chart(
                stats_df.reset_index(),
                chart_title,
                highlight_models=["o1-preview"],
            )
            barplot.figure.savefig(
                f"{stats_save_path}/performance_chart{info_type}.png"
            )
            plt.show()
            all_stats_dfs[chart_title] = stats_df# Read data from CSV file

            # Read data from CSV file
            df = pd.read_csv(f'{stats_save_path}\\final_stats.csv')

            # Sorting DataFrame by Mean Score in descending order for better visualization
            df_sorted = df.sort_values(by='mean_score', ascending=False)

            # Color palette from the provided PDF
            colors = {
                'blue_200': '#90caf9',
                'yellow_600': '#fdd835',
                'pink_200': '#f48fb1',
                'cyan_200': '#80deea',
                'orange_400': '#ffa726',
                'deep_purple_A100': '#b388ff',
                'red_700': '#d32f2f'
            }

            # Horizontal Bar Chart for Mean Score, CI Lower, and CI Upper for Each Model (Sorted in Descending Order)
            y = np.arange(len(df_sorted['model']))  # the label locations
            height = 0.25  # the height of the bars

            fig, ax = plt.subplots(figsize=(14, 10))
            bars1 = ax.barh(y - height, df_sorted['mean_score'], height, label='Mean Score', color=colors['blue_200'])
            bars2 = ax.barh(y, df_sorted['ci_lower'], height, label='CI Lower', color=colors['yellow_600'])
            bars3 = ax.barh(y + height, df_sorted['ci_upper'], height, label='CI Upper', color=colors['cyan_200'])

            # Adding labels and title
            ax.set_yticks(y)
            ax.set_yticklabels(df_sorted['model'])  # Labels on the left
            ax.set_xlabel('Scores')
            ax.set_title('Comparison of Mean Score, CI Lower, and CI Upper for Each Model')
            ax.invert_yaxis()  # Higher values at the top
            ax.legend()

            plt.tight_layout()
            plt.grid(axis='x', linestyle='--', alpha=0.5)
            plt.show()

            # Horizontal Bar Chart for Z Interval Error for Each Model (Sorted in Descending Order)
            fig, ax = plt.subplots(figsize=(14, 10))
            bars = ax.barh(df_sorted['model'], df_sorted['z_interval_error'], color=colors['pink_200'])

            plt.ylabel('Models')
            plt.xlabel('Z Interval Error')
            plt.title('Z Interval Error for Each Model')
            ax.invert_yaxis()  # Higher values at the top
            plt.tight_layout()
            plt.grid(axis='x', linestyle='--', alpha=0.5)
            plt.show()

            # Horizontal Bar Chart for Mean Score of Each Model (Sorted in Descending Order)
            fig, ax = plt.subplots(figsize=(14, 10))
            bars = ax.barh(df_sorted['model'], df_sorted['mean_score'], color=colors['orange_400'])

            plt.ylabel('Models')
            plt.xlabel('Mean Score')
            plt.title('Mean Score for Each Model')
            ax.invert_yaxis()  # Higher values at the top
            plt.tight_layout()
            plt.grid(axis='x', linestyle='--', alpha=0.5)
            plt.show()

            # Plotting Mean Score with Error Bars for Confidence Intervals (Sorted in Descending Order)
            ci_error = (df_sorted['ci_upper'] - df_sorted['ci_lower']).abs() / 2
            plt.figure(figsize=(14, 10))
            plt.errorbar(df_sorted['mean_score'], df_sorted['model'], 
                        xerr=ci_error, 
                        fmt='o', ecolor=colors['red_700'], capsize=5, label='Mean Score with CI')
            plt.ylabel('Models')
            plt.xlabel('Mean Score')
            plt.title('Mean Score with Confidence Intervals for Various Models')
            plt.gca().invert_yaxis()  # Higher values at the top
            plt.tight_layout()
            plt.legend()
            plt.grid(axis='x', linestyle='--', alpha=0.5)
            plt.show()

            # Bar Chart of Standard Deviations for Each Model

            # Create a bar chart where each model is represented individually
            fig, ax = plt.subplots(figsize=(12, 6))

            # Plotting standard deviation scores for each model
            ax.bar(df['model'], df['std_dev_score'], color='#90caf9', edgecolor='black')

            # Adding labels and title
            plt.xticks(rotation=45, ha='right', fontsize=8)
            plt.xlabel('Model')
            plt.ylabel('Standard Deviation')
            plt.title('Standard Deviation for Each Model')
            plt.tight_layout()
            plt.grid(axis='y', linestyle='--', alpha=0.5)

            plt.show()
       
            # Base directory containing rounds (e.g., 'auto_eval_save_path')
            auto_eval_save_path = auto_eval_save_path

            # Directory to save the output files
            charts_dir = stats_save_path

            os.makedirs(charts_dir, exist_ok=True)

            # Number of rounds
            answer_rounds = 2  # Update as needed

            # Function to collect all JSON file paths in a directory
            def collect_json_files(directory):
                return [os.path.join(directory, file) for file in os.listdir(directory) if file.endswith('.json')]

            # Function to process JSON files
            def process_json_files(file_paths):
                results = []
                for file_path in file_paths:
                    with open(file_path, 'r', encoding='utf-8') as file:
                        data = json.load(file)
                    for _, entry in data.items():
                        # Token calculation for each category
                        question_tokens = count_tokens(entry.get("question", ""))
                        human_answer_tokens = count_tokens(entry.get("human_answer", ""))
                        model_answer_input_tokens = count_tokens(entry.get("model_answer", ""))
                        eval_response_tokens = count_tokens(entry.get("eval_response", ""))
                        score_tokens = count_tokens(str(entry.get("score", "")))
                        bernard_evaluator_response_tokens = count_tokens(entry.get("bernard_evaluator_response", ""))
                        
                        results.append({
                            "question_tokens": question_tokens,
                            "human_answer_tokens": human_answer_tokens,
                            "model_answer_input_tokens": model_answer_input_tokens,
                            "eval_response_tokens": eval_response_tokens,
                            "score_tokens": score_tokens,
                            "bernard_evaluator_response_tokens": bernard_evaluator_response_tokens,
                            "total_tokens": question_tokens + human_answer_tokens + model_answer_input_tokens +
                                            eval_response_tokens + score_tokens + bernard_evaluator_response_tokens
                        })
                return results

            # Function to calculate tokens based on the rule: 1 token = 4 characters
            def count_tokens(text):
                return max(1, len(text) // 4)

            # Process files for each round
            all_results = []

            for round_num in range(1, answer_rounds + 1):
                round_dir = os.path.join(auto_eval_save_path, f'round_{round_num}')
                
                # Collect files from the round
                json_files = collect_json_files(round_dir)

                # Process the files in the round
                all_results.extend(process_json_files(json_files))

                # Convert results to DataFrame for analysis
                df = pd.DataFrame(all_results)

                # Summarize total tokens per category for comparison
                summary = df.sum()

                # Create a token usage comparison DataFrame
                categories = ["Question", "Human Answer", "Student Response", "Eval Response", "Score",  "Total"]
                token_usage = [
                    summary["question_tokens"],
                    summary["human_answer_tokens"],
                    summary["model_answer_input_tokens"],
                    summary["eval_response_tokens"],
                    summary["score_tokens"],
                    summary["total_tokens"]
                ]

                # Create a DataFrame for the results
                usage_df = pd.DataFrame({
                    "Category": categories,
                    "Token Usage": token_usage
                })

                # Save the token comparison table to a CSV file
                usage_csv_path = os.path.join(charts_dir, 'token_usage_comparison.csv')
                usage_df.to_csv(usage_csv_path, index=False)

                # Create a bar chart for token usage comparison
                x = np.arange(len(categories))

                # Plot Token Usage
                width = 0.35  # Width of the bars
                fig, ax = plt.subplots(figsize=(12, 6))

                bars = ax.bar(x, token_usage, width, label="Token Usage", color="#4C72B0")

                # Add values above the bars
                for bar in bars:
                    ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + 5, f"{int(bar.get_height())}", ha="center", fontsize=10)

                # Adjust the y-axis dynamically
                max_value = max(token_usage)
                ax.set_ylim(0, max_value * 1.2)  # Add 20% headroom above tallest bar

                # Add labels, title, and legend
                ax.set_ylabel("Token Count (Approx)", fontsize=12)
                ax.set_title("Token Usage Comparison for Question-Answer Pairs", fontsize=14)
                ax.set_xticks(x)
                ax.set_xticklabels(categories)
                ax.legend()

                # Save the chart as a PNG file
                chart_path = os.path.join(charts_dir, 'token_usage_comparison_chart.png')
                plt.savefig(chart_path, bbox_inches='tight')
                plt.show()


import os
import json
from loguru import logger
import requests
from openai import OpenAI
from groq import Groq
from dotenv import load_dotenv

load_dotenv()


def call_groq_evaluator_api(evaluator_model, student_answer, evaluator_system_prompt):
    api_key = os.environ.get("GROQ_API_KEY")
    gorq_api = Groq(api_key=api_key)
    completion_eval = gorq_api.chat.completions.create(
        temperature=0,
        model=evaluator_model,
        messages=evaluator_system_prompt,
    )
    response_eval = completion_eval.choices[0].message.content

    if response_eval:
        logger.info(f"call_groq_evaluator_api: {response_eval}")

        return student_answer, response_eval
    else:
        logger.error("Failed to get evaluator response.")
        return None


def call_ollama_evaluator_api(evaluator_model, student_answer, evaluator_system_prompt):
    url = "http://localhost:11434/api/chat"
    payload = {"model": evaluator_model, "messages": evaluator_system_prompt}

    # Make a single POST request (remove the duplicate)
    response = requests.post(
        url,
        json=payload,
        headers={"Content-Type": "application/json"},
        stream=True,
    )

    complete_message = ""

    # Read the streamed response line by line
    for line in response.iter_lines():
        if line:
            chunk = json.loads(line.decode("utf-8"))
            # Safely retrieve content if present
            if "message" in chunk and "content" in chunk["message"]:
                complete_message += chunk["message"]["content"]

            # If the API signals completion
            if chunk.get("done"):
                break

    print("Complete message:", complete_message)

    # Return the result (simply return the text)
    return student_answer, complete_message


def call_openrouter_student_api(full_prompt_student, warning_prompt, model_path):
    api_key = os.environ.get("OPENROUTER_KEY")
    client = OpenAI(
        base_url="https://openrouter.ai/api/v1",
        api_key=api_key,
    )
    model_path = str(model_path)

    # Make the API call
    completion = client.chat.completions.create(
        model=model_path,
        messages=[
            {
                "role": "system",
                "content": (
                    "You are a student who is being tested, please follow the directions given exactly. "
                    "You are welcomed to reason through the question. "
                    "You must return only your final answer in a JSON Object example  {'student_answer':'<My final Answer here>'}"
                ),
            },
            {
                "role": "user",
                "content": warning_prompt + full_prompt_student,
            },
        ],
    )
    # last_api_call_time = time.time()  # Update the time of the last API call
    response = completion.choices[0].message.content
    return response


def call_ollama_student_api(full_prompt_student, warning_prompt, student_model):
    url = "http://localhost:11434/api/chat"
    payload = {
        "model": student_model,
        "messages": [
            {
                "role": "system",
                "content": "You are a student who is being tested, please follow the directions given exactly. You are welcomed to reason through the question "
                + "You must return only your final answer in a JSON Object example {'student_answer':'<My final Answer here>'}",
            },
            {"role": "user", "content": warning_prompt + full_prompt_student},
        ],
    }

    response = requests.post(
        url,
        json=payload,
        headers={"Content-Type": "application/json"},
        stream=True,
    )
    complete_message = ""
    for line in response.iter_lines():
        if line:
            chunk = json.loads(line.decode("utf-8"))
            complete_message += chunk["message"]["content"]
            if chunk.get("done"):
                break
    print("line 119, LLMSCALLS ollama student = ", complete_message)
    if complete_message is not None:
        # return {"student_answer": complete_message}
        return complete_message

def call_groq_student_api(full_prompt_student, warning_prompt, groq_student_model):
    api_key = os.environ.get("GROQ_API_KEY")
    gorq_api = Groq(api_key=api_key)
    completion_eval = gorq_api.chat.completions.create(
        temperature=0,
        model=groq_student_model,
        messages=[
            {
                "role": "system",
                "content": "You are a student who is being tested, please follow the directions given exactly. You are welcomed to reason through the question "
                + "You must return only your final answer in a JSON Object example {'student_answer':'<My final Answer here>'}",
            },
            {"role": "user", "content": warning_prompt + full_prompt_student},
        ],
    )
    response = completion_eval.choices[0].message.content

    if response:
        logger.info(f"call_groq_student_api: {response}")

        return response
    else:
        logger.error("Failed to get evaluator response.")
        return None
